/*
 * cmd-list.c
 *	Functions for listing things.
 *
 * Copyright (C) 1993	Lysator computer Club,
 *			Linkoping University,  Sweden
 *
 *
 *  Everyone is granted permission to copy, modify and redistribute
 *  this code, provided the people they give it to can.
 *
 *
 * Author:	Linus Tolke
 *		Lysator Computer Club
 *		Linkoping University
 *		Sweden
 *
 * email: 	linus@lysator.liu.se
 *
 *
 *  Any opinions expressed in this code are the author's PERSONAL
 *  opinions, and does NOT represent any official standpoint of
 *  Lysator, even if so stated.
 */


#include <kom-types.h>
#include <zmalloc.h>
#include <s-string.h>

#include <kom-errno.h>
#include <services.h>

#include "xoutput.h"

#include "copy.h"
#include "error.h"
#include "cache.h"
#include "quit.h"

#include "texts.h"
#include "internal.h"

#include "cmds.h"

/* Comparing function. */
static int time_before_time (struct tm a, struct tm b)
{	/* Returns > 0 if a is a time before b */
        /* Returns < 0 if a is a time after b */
    /* The granularity is 1 second. */
    if (a.tm_year > b.tm_year)
        return 10;
    if (a.tm_year < b.tm_year)
        return -10;
    if (a.tm_yday > b.tm_yday)
        return 5;
    if (a.tm_yday < b.tm_yday)
        return -5;
    if (a.tm_hour > b.tm_hour)
        return 3;
    if (a.tm_hour < b.tm_hour)
        return -3;
    if (a.tm_min > b.tm_min)
        return 2;
    if (a.tm_min < b.tm_min)
        return -2;
    if (a.tm_sec > b.tm_sec)
        return 1;
    if (a.tm_sec < b.tm_sec)
        return -1;
    return 0;
}

struct cle {
    Conf_no 	cno;
    Conference * info;
};

static int most_recent_change (struct cle *a, struct cle *b)
{
    return time_before_time(b->info->last_written,a->info->last_written);
}

/* 
 * List confs
 * Confs are listed in "most recently recieved article"-order.
 * 
 * The problem is thus: 
 *	- Retrieve all confs.
 *	- Sort in "most recently recieved article"-order.
 * 	- Print the confs
 */

Success
cmd_list_confs (String	argument)
{
#define FUNCTION "cmd_list_confs()"

    Conf_list		  all_confs	= EMPTY_CONF_LIST;
    Success		  retval;
    struct cle		* conf_list = NULL; /* May be clobbered by longjmp */
    int			  count = 0;	    /* May be clobbered by longjmp */
    int 		  i;
    Kom_err		  error;

#define CLEAN_UP()	do {\
	release_conf_list ( &all_confs );\
	while (count > 0) { count --; release_conf(conf_list[count].info); }\
        if (conf_list) zfree(conf_list);\
			} while (0)

    TOPLOOP_SETJMP();

    CHECK_NO_ARGUMENTS();

    retval = kom_lookup_name (s_fcrea_str(""), &all_confs);
    if (retval == FAILURE)
    {
	switch (kom_errno)
	{
	case  KOM_OUT_OF_MEMORY:
	    fatal1 (CLIENT_OUT_OF_MEMORY, "lookup_name() failed");
	    break;

	case  KOM_NO_CONNECT:
	    fatal1 (CLIENT_CONNECTION_LOST, "lookup_name() failed");
	    break;

	default:
	    fatal2 (CLIENT_UNEXPECTED_SERVER_ERROR, "lookup_name() failed: %d",
		    kom_errno);
	    break;
	}
    }

    newline();

    /* all_confs does now contain a list of all confs. */
    /* Lets fetch them all in order to sort them. */
    /* We will here simultaneously map all confs to just preserve the confs. */
    
    /* Lets malloc just a little bit too much */
    /* We could count the confs first, this allocates number of pers in conf
       system to many pointers. */
    conf_list = (struct cle *) zmalloc(all_confs.no_of_confs 
				      * sizeof(struct cle));
    for (i=0; i < all_confs.no_of_confs; i++)
    {
	if (!all_confs.confs[i].type.letter_box)
	{
	    /* BUG: doesn't look for error */
	    conf_list[count].info = conf_stat(all_confs.confs[i].conf_no, &error);
	    conf_list[count].cno = all_confs.confs[i].conf_no;
	    /* secret confs are represented by NULL - pointer. */
	    /* Since we cannot sort them (we don't know when they last
	       recieved an article we leave them here. */
	    if (conf_list[count].info) /* Skipping secret confs. */
	    {
		count++;
	    }
	}
    }

    newline();

    /* Sorting. */
    qsort (conf_list, count, sizeof (struct cle), &most_recent_change);

    /* Print header line */
    xprintf(txts.list_conf_header);
    /* Lets print what we have got. */
    for (i = 0; i < count; i++)
    {
	/* Print last written time. */
	print_time(&conf_list[i].info->last_written);

	/* Print number of members. */
	xprintf(txts.list_conf_members, conf_list[i].info->no_of_members);

	/* Print number of articles. */
	xprintf(txts.list_conf_unread, conf_list[i].info->first_local_no + 
		conf_list[i].info->no_of_texts - 1);

	/* Print optional * for not member */
	if (locate_membership (conf_list[i].cno, cpn, NULL) == -1)
	{
	    xprintf(txts.list_conf_is_not_member);
	} else {
	    xprintf(txts.list_conf_is_member);
	}

	/* Print optional O for is supervisor. */
	/* Should really search the conf for membership. */
	if (cpn == conf_list[i].info->supervisor)
	{
	    xprintf(txts.list_conf_is_supervisor);
	} else {
	    xprintf(txts.list_conf_not_super);
	}

	/* Print conf name. */
	xprintf("%S\n",conf_list[i].info->name);
    }
    RETURN (OK);
}
